import graphene
from dagster._core.definitions.selector import ScheduleSelector
from dagster._core.errors import DagsterInvariantViolationError
from dagster._core.remote_representation.external import CompoundID
from dagster._core.workspace.permissions import Permissions

from dagster_graphql.implementation.fetch_schedules import (
    reset_schedule,
    start_schedule,
    stop_schedule,
)
from dagster_graphql.implementation.utils import (
    assert_permission_for_schedule,
    capture_error,
    require_permission_check,
)
from dagster_graphql.schema.errors import (
    GraphenePythonError,
    GrapheneSchedulerNotDefinedError,
    GrapheneUnauthorizedError,
)
from dagster_graphql.schema.inputs import GrapheneScheduleSelector
from dagster_graphql.schema.instigation import GrapheneInstigationState
from dagster_graphql.schema.schedules.schedules import (
    GrapheneSchedule,
    GrapheneScheduleNotFoundError,
    GrapheneScheduleOrError,
    GrapheneSchedules,
    GrapheneSchedulesOrError,
)
from dagster_graphql.schema.schedules.ticks import GrapheneInstigationTickStatus
from dagster_graphql.schema.util import ResolveInfo


class GrapheneScheduleStatus(graphene.Enum):
    RUNNING = "RUNNING"
    STOPPED = "STOPPED"
    ENDED = "ENDED"

    class Meta:
        name = "ScheduleStatus"


class GrapheneScheduler(graphene.ObjectType):
    scheduler_class = graphene.String()

    class Meta:
        name = "Scheduler"


class GrapheneSchedulerOrError(graphene.Union):
    class Meta:
        types = (GrapheneScheduler, GrapheneSchedulerNotDefinedError, GraphenePythonError)
        name = "SchedulerOrError"


class GrapheneScheduleStateResult(graphene.ObjectType):
    scheduleState = graphene.NonNull(GrapheneInstigationState)

    class Meta:
        name = "ScheduleStateResult"


class GrapheneScheduleMutationResult(graphene.Union):
    class Meta:
        types = (
            GraphenePythonError,
            GrapheneUnauthorizedError,
            GrapheneScheduleStateResult,
            GrapheneScheduleNotFoundError,
        )
        name = "ScheduleMutationResult"


class GrapheneStartScheduleMutation(graphene.Mutation):
    """Enable a schedule to launch runs for a job at a fixed interval."""

    Output = graphene.NonNull(GrapheneScheduleMutationResult)

    class Arguments:
        schedule_selector = graphene.NonNull(GrapheneScheduleSelector)

    class Meta:
        name = "StartScheduleMutation"

    @capture_error
    @require_permission_check(Permissions.START_SCHEDULE)
    def mutate(self, graphene_info: ResolveInfo, schedule_selector):
        selector = ScheduleSelector.from_graphql_input(schedule_selector)
        assert_permission_for_schedule(graphene_info, Permissions.START_SCHEDULE, selector)
        return start_schedule(graphene_info, selector)


class GrapheneStopRunningScheduleMutation(graphene.Mutation):
    """Disable a schedule from launching runs for a job."""

    Output = graphene.NonNull(GrapheneScheduleMutationResult)

    class Arguments:
        id = graphene.Argument(graphene.String)  # Schedule / InstigationState id
        schedule_origin_id = graphene.Argument(graphene.String)
        schedule_selector_id = graphene.Argument(graphene.String)

    class Meta:
        name = "StopRunningScheduleMutation"

    @capture_error
    @require_permission_check(Permissions.STOP_RUNNING_SCHEDULE)
    def mutate(
        self,
        graphene_info: ResolveInfo,
        id: str | None = None,
        schedule_origin_id: str | None = None,
        schedule_selector_id: str | None = None,
    ):
        if id:
            cid = CompoundID.from_string(id)
            schedule_origin_id = cid.remote_origin_id
            schedule_selector_id = cid.selector_id
        elif schedule_origin_id and CompoundID.is_valid_string(schedule_origin_id):
            # cross-push handle if InstigationState.id being passed through as origin id
            cid = CompoundID.from_string(schedule_origin_id)
            schedule_origin_id = cid.remote_origin_id
            schedule_selector_id = cid.selector_id
        elif schedule_origin_id is None or schedule_selector_id is None:
            raise DagsterInvariantViolationError(
                "Must specify id or scheduleOriginId and scheduleSelectorId"
            )

        return stop_schedule(graphene_info, schedule_origin_id, schedule_selector_id)


class GrapheneResetScheduleMutation(graphene.Mutation):
    """Reset a schedule to its status defined in code, otherwise disable it from launching runs for a job."""

    Output = graphene.NonNull(GrapheneScheduleMutationResult)

    class Arguments:
        schedule_selector = graphene.NonNull(GrapheneScheduleSelector)

    class Meta:
        name = "ResetScheduleMutation"

    @capture_error
    @require_permission_check(Permissions.START_SCHEDULE)
    @require_permission_check(Permissions.STOP_RUNNING_SCHEDULE)
    def mutate(self, graphene_info: ResolveInfo, schedule_selector):
        selector = ScheduleSelector.from_graphql_input(schedule_selector)

        assert_permission_for_schedule(graphene_info, Permissions.START_SCHEDULE, selector)
        assert_permission_for_schedule(graphene_info, Permissions.STOP_RUNNING_SCHEDULE, selector)

        return reset_schedule(graphene_info, selector)


def types():
    from dagster_graphql.schema.schedules.ticks import (
        GrapheneScheduleTick,
        GrapheneScheduleTickFailureData,
        GrapheneScheduleTickSpecificData,
        GrapheneScheduleTickSuccessData,
    )

    # Double check mutations don't appear twice
    return [
        GrapheneInstigationTickStatus,
        GrapheneSchedule,
        GrapheneScheduleMutationResult,
        GrapheneScheduleOrError,
        GrapheneScheduler,
        GrapheneSchedulerOrError,
        GrapheneSchedules,
        GrapheneSchedulesOrError,
        GrapheneScheduleStateResult,
        GrapheneScheduleStatus,
        GrapheneScheduleTick,
        GrapheneScheduleTickFailureData,
        GrapheneScheduleTickSpecificData,
        GrapheneScheduleTickSuccessData,
        GrapheneStartScheduleMutation,
        GrapheneStopRunningScheduleMutation,
        GrapheneResetScheduleMutation,
    ]



from typing import Any

from dagster._core.utils import make_new_run_id
from dagster_graphql.test.utils import execute_dagster_graphql

from dagster_graphql_tests.graphql.graphql_context_test_suite import (
    GraphQLContextVariant,
    make_graphql_context_test_suite,
)

INSTANCE_QUERY = """
query InstanceDetailSummaryQuery {
    instance {
        runQueuingSupported
        hasInfo
        useAutoMaterializeSensors
        poolConfig {
            poolGranularity
            defaultPoolLimit
            opGranularityRunBuffer
        }
    }
}
"""

GET_CONCURRENCY_LIMITS_QUERY = """
query InstanceConcurrencyLimitsQuery($concurrencyKey: String!) {
    instance {
        concurrencyLimit(concurrencyKey: $concurrencyKey) {
            concurrencyKey
            slotCount
            activeSlotCount
            activeRunIds
            claimedSlots {
                runId
                stepKey
            }
            pendingSteps {
                runId
                stepKey
                enqueuedTimestamp
                assignedTimestamp
                priority
            }
            limit
            usingDefaultLimit
        }
    }
}

"""

ALL_CONCURRENCY_LIMITS_QUERY = """
query AllConcurrencyLimitsQuery {
    instance {
        concurrencyLimits {
            concurrencyKey
            slotCount
            activeSlotCount
            activeRunIds
            claimedSlots {
                runId
                stepKey
            }
            pendingSteps {
                runId
                stepKey
                enqueuedTimestamp
                assignedTimestamp
                priority
            }
            limit
            usingDefaultLimit
        }
    }
}
"""

SET_CONCURRENCY_LIMITS_MUTATION = """
mutation SetConcurrencyLimit($concurrencyKey: String!, $limit: Int!) {
    setConcurrencyLimit(concurrencyKey: $concurrencyKey, limit: $limit)
}
"""

FREE_CONCURRENCY_SLOTS_FOR_RUN_MUTATION = """
mutation FreeConcurrencySlotsForRun($runId: String!) {
    freeConcurrencySlotsForRun(runId: $runId)
}
"""

FREE_CONCURRENCY_SLOTS_MUTATION = """
mutation FreeConcurrencySlots($runId: String!, $stepKey: String) {
    freeConcurrencySlots(runId: $runId, stepKey: $stepKey)
}
"""

BaseTestSuite: Any = make_graphql_context_test_suite(
    context_variants=[
        GraphQLContextVariant.sqlite_with_queued_run_coordinator_managed_grpc_env(),
    ]
)


def fetch_concurrency_limit(graphql_context, key: str):
    results = execute_dagster_graphql(
        graphql_context,
        GET_CONCURRENCY_LIMITS_QUERY,
        {"concurrencyKey": key},
    )
    assert results.data
    assert "instance" in results.data
    assert "concurrencyLimit" in results.data["instance"]
    return results.data["instance"]["concurrencyLimit"]


def set_concurrency_limit(graphql_context, key: str, limit: int):
    execute_dagster_graphql(
        graphql_context,
        SET_CONCURRENCY_LIMITS_MUTATION,
        variables={
            "concurrencyKey": key,
            "limit": limit,
        },
    )


def fetch_all_concurrency_limits(graphql_context):
    results = execute_dagster_graphql(
        graphql_context,
        ALL_CONCURRENCY_LIMITS_QUERY,
    )
    assert results.data
    assert "instance" in results.data
    assert "concurrencyLimits" in results.data["instance"]
    return [limit for limit in results.data["instance"]["concurrencyLimits"]]


class TestInstanceSettings(BaseTestSuite):
    def test_instance_settings(self, graphql_context):
        results = execute_dagster_graphql(graphql_context, INSTANCE_QUERY)
        assert results.data == {
            "instance": {
                "runQueuingSupported": True,
                "hasInfo": graphql_context.show_instance_config,
                "useAutoMaterializeSensors": graphql_context.instance.auto_materialize_use_sensors,
                "poolConfig": {
                    "poolGranularity": None,
                    "defaultPoolLimit": None,
                    "opGranularityRunBuffer": None,
                },
            }
        }

    def test_concurrency_limits(self, graphql_context):
        instance = graphql_context.instance

        # default limits are empty
        all_limits = fetch_all_concurrency_limits(graphql_context)
        assert len(all_limits) == 0

        # set a limit
        set_concurrency_limit(graphql_context, "foo", 10)
        foo = fetch_concurrency_limit(graphql_context, "foo")
        assert foo["concurrencyKey"] == "foo"
        assert foo["slotCount"] == 10
        assert foo["activeSlotCount"] == 0
        assert foo["activeRunIds"] == []
        assert foo["claimedSlots"] == []
        assert foo["pendingSteps"] == []

        # claim a slot
        run_id = make_new_run_id()
        instance.event_log_storage.claim_concurrency_slot("foo", run_id, "fake_step_key")
        foo = fetch_concurrency_limit(graphql_context, "foo")
        assert foo["concurrencyKey"] == "foo"
        assert foo["slotCount"] == 10
        assert foo["activeSlotCount"] == 1
        assert foo["activeRunIds"] == [run_id]
        assert foo["claimedSlots"] == [{"runId": run_id, "stepKey": "fake_step_key"}]
        assert len(foo["pendingSteps"]) == 1
        assert foo["pendingSteps"][0]["runId"] == run_id
        assert foo["pendingSteps"][0]["stepKey"] == "fake_step_key"
        assert foo["pendingSteps"][0]["assignedTimestamp"] is not None
        assert foo["pendingSteps"][0]["priority"] == 0

        set_concurrency_limit(graphql_context, "foo", 5)
        foo = fetch_concurrency_limit(graphql_context, "foo")
        assert foo["concurrencyKey"] == "foo"
        assert foo["slotCount"] == 5
        assert foo["activeSlotCount"] == 1
        assert foo["activeRunIds"] == [run_id]
        assert foo["claimedSlots"] == [{"runId": run_id, "stepKey": "fake_step_key"}]
        assert len(foo["pendingSteps"]) == 1
        assert foo["pendingSteps"][0]["runId"] == run_id
        assert foo["pendingSteps"][0]["stepKey"] == "fake_step_key"
        assert foo["pendingSteps"][0]["assignedTimestamp"] is not None
        assert foo["pendingSteps"][0]["priority"] == 0

        instance.event_log_storage.free_concurrency_slots_for_run(run_id)
        foo = fetch_concurrency_limit(graphql_context, "foo")
        assert foo["concurrencyKey"] == "foo"
        assert foo["slotCount"] == 5
        assert foo["activeSlotCount"] == 0
        assert foo["activeRunIds"] == []
        assert foo["claimedSlots"] == []
        assert foo["pendingSteps"] == []

    def test_concurrency_free(self, graphql_context):
        storage = graphql_context.instance.event_log_storage

        # set a limit
        storage.set_concurrency_slots("foo", 1)

        # claim the slot
        run_id = make_new_run_id()
        run_id_2 = make_new_run_id()
        storage.claim_concurrency_slot("foo", run_id, "fake_step_key")
        # add pending steps
        storage.claim_concurrency_slot("foo", run_id, "fake_step_key_2")
        storage.claim_concurrency_slot("foo", run_id_2, "fake_step_key_3")

        foo_info = storage.get_concurrency_info("foo")
        assert foo_info.slot_count == 1
        assert foo_info.active_slot_count == 1
        assert foo_info.active_run_ids == {run_id}
        assert foo_info.pending_step_count == 2
        assert foo_info.pending_run_ids == {run_id, run_id_2}
        assert foo_info.assigned_step_count == 1
        assert foo_info.assigned_run_ids == {run_id}

        execute_dagster_graphql(
            graphql_context,
            FREE_CONCURRENCY_SLOTS_MUTATION,
            variables={"runId": run_id_2, "stepKey": "fake_step_key_3"},
        )

        foo_info = storage.get_concurrency_info("foo")
        assert foo_info.slot_count == 1
        assert foo_info.active_slot_count == 1
        assert foo_info.active_run_ids == {run_id}
        assert foo_info.pending_step_count == 1
        assert foo_info.pending_run_ids == {run_id}
        assert foo_info.assigned_step_count == 1
        assert foo_info.assigned_run_ids == {run_id}

        execute_dagster_graphql(
            graphql_context,
            FREE_CONCURRENCY_SLOTS_MUTATION,
            variables={"runId": run_id},
        )
        foo_info = storage.get_concurrency_info("foo")
        assert foo_info.slot_count == 1
        assert foo_info.active_slot_count == 0
        assert foo_info.active_run_ids == set()
        assert foo_info.pending_step_count == 0
        assert foo_info.pending_run_ids == set()
        assert foo_info.assigned_step_count == 0
        assert foo_info.assigned_run_ids == set()

    def test_concurrency_free_run(self, graphql_context):
        storage = graphql_context.instance.event_log_storage

        # set a limit
        storage.set_concurrency_slots("foo", 1)

        # claim the slot
        run_id = make_new_run_id()
        run_id_2 = make_new_run_id()
        storage.claim_concurrency_slot("foo", run_id, "fake_step_key")
        # add pending steps
        storage.claim_concurrency_slot("foo", run_id, "fake_step_key_2")
        storage.claim_concurrency_slot("foo", run_id_2, "fake_step_key_3")

        foo_info = storage.get_concurrency_info("foo")
        assert foo_info.slot_count == 1
        assert foo_info.active_slot_count == 1
        assert foo_info.active_run_ids == {run_id}
        assert foo_info.pending_step_count == 2
        assert foo_info.pending_run_ids == {run_id, run_id_2}
        assert foo_info.assigned_step_count == 1
        assert foo_info.assigned_run_ids == {run_id}

        execute_dagster_graphql(
            graphql_context,
            FREE_CONCURRENCY_SLOTS_FOR_RUN_MUTATION,
            variables={"runId": run_id},
        )

        foo_info = storage.get_concurrency_info("foo")
        assert foo_info.slot_count == 1
        assert foo_info.active_slot_count == 0
        assert foo_info.active_run_ids == set()
        assert foo_info.pending_step_count == 0
        assert foo_info.pending_run_ids == set()
        assert foo_info.assigned_step_count == 1
        assert foo_info.assigned_run_ids == {run_id_2}


ConcurrencyTestSuite: Any = make_graphql_context_test_suite(
    context_variants=[
        GraphQLContextVariant.sqlite_with_default_concurrency_managed_grpc_env(),
    ]
)


class TestConcurrencyInstanceSettings(ConcurrencyTestSuite):
    def test_default_concurrency(self, graphql_context):
        # no limits
        all_limits = fetch_all_concurrency_limits(graphql_context)
        assert len(all_limits) == 0

        # default limits are empty
        limit = fetch_concurrency_limit(graphql_context, "foo")
        assert limit is not None
        assert limit["slotCount"] == 0
        assert limit["limit"] == 1
        assert limit["usingDefaultLimit"]

        # set a limit
        set_concurrency_limit(graphql_context, "foo", 0)

        limit = fetch_concurrency_limit(graphql_context, "foo")
        assert limit is not None
        assert limit["slotCount"] == 0
        assert limit["limit"] == 0
        assert not limit["usingDefaultLimit"]

        # instance settings
        results = execute_dagster_graphql(graphql_context, INSTANCE_QUERY)
        assert results.data == {
            "instance": {
                "runQueuingSupported": True,
                "hasInfo": graphql_context.show_instance_config,
                "useAutoMaterializeSensors": graphql_context.instance.auto_materialize_use_sensors,
                "poolConfig": {
                    "poolGranularity": None,
                    "defaultPoolLimit": 1,
                    "opGranularityRunBuffer": None,
                },
            }
        }



from collections.abc import Mapping, Sequence
from typing import TYPE_CHECKING, cast

from dagster import _check as check
from dagster._core.definitions.asset_checks.asset_check_spec import AssetCheckKey
from dagster._core.definitions.partitions.definition import (
    MultiPartitionsDefinition,
    PartitionsDefinition,
    TimeWindowPartitionsDefinition,
)
from dagster._core.definitions.partitions.subset import PartitionsSubset, TimeWindowPartitionsSubset
from dagster._core.event_api import PartitionKeyFilter
from dagster._core.events import ASSET_CHECK_EVENTS, DagsterEventType
from dagster._core.instance import DynamicPartitionsStore
from dagster._core.loader import LoadingContext
from dagster._core.remote_representation.handle import RepositoryHandle
from dagster._core.storage.asset_check_execution_record import (
    AssetCheckExecutionRecordStatus,
    AssetCheckExecutionResolvedStatus,
    AssetCheckInstanceSupport,
)
from dagster._core.storage.asset_check_state import AssetCheckState
from dagster._core.storage.dagster_run import RunRecord
from packaging import version

from dagster_graphql.implementation.partition_status_utils import (
    build_multi_partition_ranges_generic,
    build_time_partition_ranges_generic,
    extract_partition_keys_by_status,
)
from dagster_graphql.schema.asset_checks import (
    AssetCheckPartitionRangeStatus,
    GrapheneAssetCheckDefaultPartitionStatuses,
    GrapheneAssetCheckMultiPartitionRangeStatuses,
    GrapheneAssetCheckMultiPartitionStatuses,
    GrapheneAssetCheckNeedsAgentUpgradeError,
    GrapheneAssetCheckNeedsMigrationError,
    GrapheneAssetCheckNeedsUserCodeUpgrade,
    GrapheneAssetCheckTimePartitionRangeStatus,
    GrapheneAssetCheckTimePartitionStatuses,
)

if TYPE_CHECKING:
    from dagster_graphql.schema.asset_checks import GrapheneAssetCheckExecution
    from dagster_graphql.schema.entity_key import GrapheneAssetCheckHandle
    from dagster_graphql.schema.util import ResolveInfo


def check_asset_checks_support(
    graphene_info: "ResolveInfo", repository_handle: RepositoryHandle
) -> (
    GrapheneAssetCheckNeedsMigrationError
    | GrapheneAssetCheckNeedsUserCodeUpgrade
    | GrapheneAssetCheckNeedsAgentUpgradeError
    | None
):
    asset_check_support = graphene_info.context.instance.get_asset_check_support()
    if asset_check_support == AssetCheckInstanceSupport.NEEDS_MIGRATION:
        return GrapheneAssetCheckNeedsMigrationError(
            message="Asset checks require an instance migration. Run `dagster instance migrate`."
        )
    elif asset_check_support == AssetCheckInstanceSupport.NEEDS_AGENT_UPGRADE:
        return GrapheneAssetCheckNeedsAgentUpgradeError(
            "Asset checks require an agent upgrade to 1.5.0 or greater."
        )
    else:
        check.invariant(
            asset_check_support == AssetCheckInstanceSupport.SUPPORTED,
            f"Unexpected asset check support status {asset_check_support}",
        )

    library_versions = graphene_info.context.get_dagster_library_versions(
        repository_handle.location_name
    )
    code_location_version = (library_versions or {}).get("dagster")
    if code_location_version and version.parse(code_location_version) < version.parse("1.5"):
        return GrapheneAssetCheckNeedsUserCodeUpgrade(
            message=(
                "Asset checks require dagster>=1.5. Upgrade your dagster"
                " version for this code location."
            )
        )


def fetch_asset_check_executions(
    loading_context: LoadingContext,
    asset_check_key: AssetCheckKey,
    limit: int,
    cursor: str | None,
    partition_filter: PartitionKeyFilter | None,
) -> list["GrapheneAssetCheckExecution"]:
    from dagster_graphql.schema.asset_checks import GrapheneAssetCheckExecution

    check_records = loading_context.instance.event_log_storage.get_asset_check_execution_history(
        check_key=asset_check_key,
        limit=limit,
        cursor=int(cursor) if cursor else None,
        partition_filter=partition_filter,
    )

    RunRecord.prepare(
        loading_context,
        [r.run_id for r in check_records if r.status == AssetCheckExecutionRecordStatus.PLANNED],
    )

    return [GrapheneAssetCheckExecution(check_record) for check_record in check_records]


def get_asset_checks_for_run_id(
    graphene_info: "ResolveInfo", run_id: str
) -> Sequence["GrapheneAssetCheckHandle"]:
    from dagster_graphql.schema.entity_key import GrapheneAssetCheckHandle

    check.str_param(run_id, "run_id")

    records = graphene_info.context.instance.all_logs(run_id, of_type=ASSET_CHECK_EVENTS)

    asset_check_keys = set(
        [
            record.get_dagster_event().asset_check_evaluation_data.asset_check_key
            for record in records
            if record.is_dagster_event
            and record.get_dagster_event().event_type == DagsterEventType.ASSET_CHECK_EVALUATION
        ]
        + [
            record.get_dagster_event().asset_check_planned_data.asset_check_key
            for record in records
            if record.is_dagster_event
            and record.get_dagster_event().event_type
            == DagsterEventType.ASSET_CHECK_EVALUATION_PLANNED
        ]
    )
    return [
        GrapheneAssetCheckHandle(handle=asset_check_key)
        for asset_check_key in sorted(
            asset_check_keys, key=lambda key: key.name + "".join(key.asset_key.path)
        )
    ]


# Partition status builder functions


def build_asset_check_partition_statuses(
    key: AssetCheckKey,
    asset_check_state: AssetCheckState,
    partitions_def: PartitionsDefinition,
    dynamic_partitions_store: DynamicPartitionsStore,
) -> (
    GrapheneAssetCheckTimePartitionStatuses
    | GrapheneAssetCheckDefaultPartitionStatuses
    | GrapheneAssetCheckMultiPartitionStatuses
):
    """Build appropriate partition status type based on partitions definition.

    Decision tree:
    1. Extract subsets from partition_status.subsets for all 5 statuses
    2. Check first subset type:
       - TimeWindowPartitionsSubset → build_time_partition_statuses_for_checks()
       - MultiPartitionsDefinition in partitions_def → build_multi_partition_statuses_for_checks()
       - Otherwise → build_default_partition_statuses_for_checks()

    Note: Missing status is calculated on the frontend as the complement of all known statuses.
    """
    # Extract subsets for each status
    subsets = asset_check_state.subsets

    # Check if we have any subsets to determine the type
    first_subset = next((s.subset_value for s in subsets.values() if s is not None), None)

    # Decision based on subset type and partitions definition
    if first_subset is not None and isinstance(first_subset, TimeWindowPartitionsSubset):
        # Time partitions - use range compression
        return build_time_partition_statuses_for_checks(
            subsets={
                status: cast("TimeWindowPartitionsSubset", s.subset_value)
                for status, s in subsets.items()
                if s is not None
            },
            partitions_def=cast("TimeWindowPartitionsDefinition", partitions_def),
        )
    elif isinstance(partitions_def, MultiPartitionsDefinition):
        # Multi-partitions - use 2D run-length encoding
        return build_multi_partition_statuses_for_checks(
            key=key,
            subsets={status: s.subset_value for status, s in subsets.items() if s is not None},
            partitions_def=partitions_def,
            dynamic_partitions_store=dynamic_partitions_store,
        )
    else:
        # Default partitions - simple partition key lists
        return build_default_partition_statuses_for_checks(
            subsets={status: s.subset_value for status, s in subsets.items() if s is not None},
            partitions_def=partitions_def,
            dynamic_partitions_store=dynamic_partitions_store,
        )


def build_time_partition_statuses_for_checks(
    subsets: Mapping[AssetCheckExecutionResolvedStatus, TimeWindowPartitionsSubset],
    partitions_def: TimeWindowPartitionsDefinition,
) -> GrapheneAssetCheckTimePartitionStatuses:
    """Build time-based ranges for all 5 statuses.

    Uses shared utility to convert time window subsets to generic ranges,
    then converts to GraphQL-specific types.

    Note: Missing status is calculated on the frontend as the complement of all known statuses.
    """
    # Use shared utility to build generic ranges
    generic_ranges = build_time_partition_ranges_generic(subsets)

    # Convert generic ranges to GraphQL types
    graphene_ranges = [
        GrapheneAssetCheckTimePartitionRangeStatus(
            startTime=r.start_time,
            endTime=r.end_time,
            startKey=r.start_key,
            endKey=r.end_key,
            status=map_check_status_to_range_status(r.status),
        )
        for r in generic_ranges
    ]

    return GrapheneAssetCheckTimePartitionStatuses(ranges=graphene_ranges)


def build_default_partition_statuses_for_checks(
    subsets: Mapping[AssetCheckExecutionResolvedStatus, PartitionsSubset],
    partitions_def: PartitionsDefinition,
    dynamic_partitions_store: DynamicPartitionsStore,
) -> GrapheneAssetCheckDefaultPartitionStatuses:
    """Extract partition keys for each status.

    Uses shared utility to extract partition keys by status,
    then maps to GraphQL field names.

    Note: Missing status is calculated on the frontend as the complement of all known statuses.
    """
    # Use shared utility to extract partition keys
    keys_by_status = extract_partition_keys_by_status(
        subsets, partitions_def, dynamic_partitions_store
    )

    return GrapheneAssetCheckDefaultPartitionStatuses(
        succeededPartitions=keys_by_status.get(AssetCheckExecutionResolvedStatus.SUCCEEDED, []),
        failedPartitions=keys_by_status.get(AssetCheckExecutionResolvedStatus.FAILED, []),
        inProgressPartitions=keys_by_status.get(AssetCheckExecutionResolvedStatus.IN_PROGRESS, []),
        skippedPartitions=keys_by_status.get(AssetCheckExecutionResolvedStatus.SKIPPED, []),
        executionFailedPartitions=keys_by_status.get(
            AssetCheckExecutionResolvedStatus.EXECUTION_FAILED, []
        ),
    )


def build_multi_partition_statuses_for_checks(
    key: AssetCheckKey,
    subsets: Mapping[AssetCheckExecutionResolvedStatus, PartitionsSubset],
    partitions_def: MultiPartitionsDefinition,
    dynamic_partitions_store: DynamicPartitionsStore,
) -> GrapheneAssetCheckMultiPartitionStatuses:
    """Build 2D run-length encoded statuses.

    Uses shared utility for 2D run-length encoding algorithm,
    then converts to GraphQL-specific types.

    Note: Missing status is calculated on the frontend as the complement of all known statuses.
    """
    from dagster._core.asset_graph_view.serializable_entity_subset import SerializableEntitySubset

    # Define recursive builder for secondary dimension
    def build_secondary_dim(
        secondary_subsets: Mapping[AssetCheckExecutionResolvedStatus, PartitionsSubset],
    ):
        # Convert to SerializableEntitySubset for recursive call
        secondary_serializable_subsets = {}
        for status, secondary_subset in secondary_subsets.items():
            secondary_serializable_subsets[status] = SerializableEntitySubset.from_coercible_value(
                key,
                list(secondary_subset.get_partition_keys()),
                partitions_def.secondary_dimension.partitions_def,
            )

        # Build temporary cache value for secondary dimension
        temp_cache_value = AssetCheckState(
            latest_storage_id=0,
            subsets=secondary_serializable_subsets,
            in_progress_runs={},
        )

        # Recursively build secondary dimension statuses
        return build_asset_check_partition_statuses(
            key=key,
            asset_check_state=temp_cache_value,
            partitions_def=partitions_def.secondary_dimension.partitions_def,
            dynamic_partitions_store=dynamic_partitions_store,
        )

    # Use shared utility to build generic multi-partition ranges
    generic_ranges = build_multi_partition_ranges_generic(
        subsets,
        partitions_def,
        dynamic_partitions_store,
        build_secondary_dim,
    )

    # Convert generic ranges to GraphQL types
    graphene_ranges = [
        GrapheneAssetCheckMultiPartitionRangeStatuses(
            primaryDimStartKey=r.primary_dim_start_key,
            primaryDimEndKey=r.primary_dim_end_key,
            primaryDimStartTime=r.primary_dim_start_time,
            primaryDimEndTime=r.primary_dim_end_time,
            secondaryDim=r.secondary_dim,
        )
        for r in generic_ranges
    ]

    return GrapheneAssetCheckMultiPartitionStatuses(
        ranges=graphene_ranges,
        primaryDimensionName=partitions_def.primary_dimension.name,
    )


# Helper functions


def map_check_status_to_range_status(
    status: AssetCheckExecutionResolvedStatus,
) -> AssetCheckPartitionRangeStatus:
    """Map internal status enum to GraphQL range status enum."""
    mapping = {
        AssetCheckExecutionResolvedStatus.SUCCEEDED: AssetCheckPartitionRangeStatus.SUCCEEDED,
        AssetCheckExecutionResolvedStatus.FAILED: AssetCheckPartitionRangeStatus.FAILED,
        AssetCheckExecutionResolvedStatus.IN_PROGRESS: AssetCheckPartitionRangeStatus.IN_PROGRESS,
        AssetCheckExecutionResolvedStatus.SKIPPED: AssetCheckPartitionRangeStatus.SKIPPED,
        AssetCheckExecutionResolvedStatus.EXECUTION_FAILED: AssetCheckPartitionRangeStatus.EXECUTION_FAILED,
    }
    return mapping[status]



from dagster import AssetKey, SourceAsset, asset, repository
from dagster._core.definitions.decorators.source_asset_decorator import observable_source_asset


@asset
def derived_asset():
    return 5


@repository
def upstream_assets_repository():
    return [derived_asset]


@observable_source_asset
def sometimes_observable_source_asset():
    return 5


unexecutable_src = SourceAsset("sometimes_observable_source_asset")

source_assets = [SourceAsset(AssetKey("derived_asset")), SourceAsset("always_source_asset")]


@asset
def downstream_asset1(derived_asset, always_source_asset):
    assert derived_asset


@asset
def downstream_asset2(derived_asset, always_source_asset):
    assert derived_asset


@repository
def downstream_assets_repository1():
    return [downstream_asset1, *source_assets, unexecutable_src]


@repository
def downstream_assets_repository2():
    return [downstream_asset2, *source_assets, sometimes_observable_source_asset]



import sys
from typing import Any
from unittest import mock

from dagster import file_relative_path, repository
from dagster._core.code_pointer import CodePointer
from dagster._core.remote_origin import ManagedGrpcPythonEnvCodeLocationOrigin
from dagster._core.remote_representation.external_data import RepositorySnap
from dagster._core.types.loadable_target_origin import LoadableTargetOrigin
from dagster._core.workspace.load import location_origins_from_yaml_paths
from dagster._grpc.types import ListRepositoriesResponse
from dagster_graphql.test.utils import execute_dagster_graphql, main_repo_location_name

from dagster_graphql_tests.graphql.graphql_context_test_suite import (
    GraphQLContextVariant,
    ReadonlyGraphQLContextTestMatrix,
    make_graphql_context_test_suite,
)

RELOAD_REPOSITORY_LOCATION_QUERY = """
mutation ($repositoryLocationName: String!) {
   reloadRepositoryLocation(repositoryLocationName: $repositoryLocationName) {
      __typename
      ... on WorkspaceLocationEntry {
        id
        name
        loadStatus
        locationOrLoadError {
            __typename
            ... on RepositoryLocation {
                name
                repositories {
                    name
                    displayMetadata {
                        key
                        value
                    }
                }
                isReloadSupported
            }
            ... on PythonError {
                message
            }
        }
      }
   }
}
"""

RELOAD_WORKSPACE_QUERY = """
mutation {
   reloadWorkspace {
      __typename
      ... on Workspace {
        locationEntries {
          __typename
          id
          name
          loadStatus
          locationOrLoadError {
            __typename
            ... on RepositoryLocation {
                id
                name
                repositories {
                    name
                }
                isReloadSupported
            }
            ... on PythonError {
                message
            }
          }
        }
      }
  }
}
"""


MultiLocationTestSuite: Any = make_graphql_context_test_suite(
    context_variants=[GraphQLContextVariant.non_launchable_sqlite_instance_multi_location()]
)
OutOfProcessTestSuite: Any = make_graphql_context_test_suite(
    context_variants=[GraphQLContextVariant.non_launchable_sqlite_instance_managed_grpc_env()]
)
ManagedTestSuite: Any = make_graphql_context_test_suite(
    context_variants=[
        GraphQLContextVariant.non_launchable_sqlite_instance_managed_grpc_env(),
    ]
)
CodeServerCliTestSuite: Any = make_graphql_context_test_suite(
    context_variants=[
        GraphQLContextVariant.sqlite_with_default_run_launcher_code_server_cli_env(),
    ]
)


class TestReloadWorkspaceReadOnly(ReadonlyGraphQLContextTestMatrix):
    def test_reload_workspace_permission_failure(self, graphql_context):
        result = execute_dagster_graphql(graphql_context, RELOAD_WORKSPACE_QUERY)

        assert result
        assert result.data
        assert result.data["reloadWorkspace"]
        assert result.data["reloadWorkspace"]["__typename"] == "UnauthorizedError"


class TestReloadWorkspace(MultiLocationTestSuite):
    def test_reload_workspace(self, graphql_context):
        result = execute_dagster_graphql(graphql_context, RELOAD_WORKSPACE_QUERY)

        assert result
        assert result.data
        assert result.data["reloadWorkspace"]
        assert result.data["reloadWorkspace"]["__typename"] == "Workspace"

        nodes = result.data["reloadWorkspace"]["locationEntries"]

        assert len(nodes) == 2

        assert all(
            [node["locationOrLoadError"]["__typename"] == "RepositoryLocation" for node in nodes]
        )

        original_origins = location_origins_from_yaml_paths(
            [file_relative_path(__file__, "multi_location.yaml")]
        )

        # simulate removing all the origins
        with mock.patch(
            "dagster._core.workspace.load_target.location_origins_from_yaml_paths",
        ) as origins_mock:
            # simulate removing an origin, reload

            origins_mock.return_value = original_origins[0:1]
            result = execute_dagster_graphql(graphql_context, RELOAD_WORKSPACE_QUERY)

            assert result
            assert result.data
            assert result.data["reloadWorkspace"]
            assert result.data["reloadWorkspace"]["__typename"] == "Workspace"

            nodes = result.data["reloadWorkspace"]["locationEntries"]

            assert len(nodes) == 1

            assert all(
                [
                    node["locationOrLoadError"]["__typename"] == "RepositoryLocation"
                    and node["loadStatus"] == "LOADED"
                    for node in nodes
                ]
            )

            # Simulate adding an origin with an error, reload

            original_origins.append(  # pyright: ignore[reportAttributeAccessIssue]
                ManagedGrpcPythonEnvCodeLocationOrigin(
                    location_name="error_location",
                    loadable_target_origin=LoadableTargetOrigin(
                        python_file="made_up_file.py", executable_path=sys.executable
                    ),
                )
            )

            origins_mock.return_value = original_origins

            result = execute_dagster_graphql(graphql_context, RELOAD_WORKSPACE_QUERY)

            assert result
            assert result.data
            assert result.data["reloadWorkspace"]
            assert result.data["reloadWorkspace"]["__typename"] == "Workspace"

            nodes = result.data["reloadWorkspace"]["locationEntries"]
            assert len(nodes) == 3

            assert (
                len(
                    [
                        node
                        for node in nodes
                        if node["locationOrLoadError"]["__typename"] == "RepositoryLocation"
                        and node["loadStatus"] == "LOADED"
                    ]
                )
                == 2
            )
            failures = [
                node for node in nodes if node["locationOrLoadError"]["__typename"] == "PythonError"
            ]
            assert len(failures) == 1
            assert failures[0]["name"] == "error_location"
            assert failures[0]["loadStatus"] == "LOADED"

            # Add another origin without an error, reload

            original_origins.append(original_origins[0]._replace(location_name="location_copy"))  # pyright: ignore[reportAttributeAccessIssue]
            origins_mock.return_value = original_origins

            result = execute_dagster_graphql(graphql_context, RELOAD_WORKSPACE_QUERY)

            nodes = result.data["reloadWorkspace"]["locationEntries"]
            assert len(nodes) == 4

            assert (
                len(
                    [
                        node
                        for node in nodes
                        if node["locationOrLoadError"]["__typename"] == "RepositoryLocation"
                    ]
                )
                == 3
            )
            failures = [
                node for node in nodes if node["locationOrLoadError"]["__typename"] == "PythonError"
            ]
            assert len(failures) == 1

            assert "location_copy" in [node["name"] for node in nodes]
            assert original_origins[0].location_name in [node["name"] for node in nodes]

            # Finally, update one of the origins' location names

            original_origins[0] = original_origins[0]._replace(location_name="new_location_name")  # pyright: ignore[reportIndexIssue,reportAttributeAccessIssue]

            result = execute_dagster_graphql(graphql_context, RELOAD_WORKSPACE_QUERY)

            nodes = result.data["reloadWorkspace"]["locationEntries"]
            assert len(nodes) == 4

            assert (
                len(
                    [
                        node
                        for node in nodes
                        if node["locationOrLoadError"]["__typename"] == "RepositoryLocation"
                    ]
                )
                == 3
            )
            failures = [
                node for node in nodes if node["locationOrLoadError"]["__typename"] == "PythonError"
            ]
            assert len(failures) == 1

            assert "new_location_name" in [node["name"] for node in nodes]


class TestReloadRepositoriesReadOnly(ReadonlyGraphQLContextTestMatrix):
    def test_reload_repository_permission_failure(self, graphql_context):
        result = execute_dagster_graphql(
            graphql_context,
            RELOAD_REPOSITORY_LOCATION_QUERY,
            {"repositoryLocationName": main_repo_location_name()},
        )

        assert result
        assert result.data
        assert result.data["reloadRepositoryLocation"]
        assert result.data["reloadRepositoryLocation"]["__typename"] == "UnauthorizedError"


class TestReloadRepositoriesOutOfProcess(OutOfProcessTestSuite):
    def test_out_of_process_reload_location(self, graphql_context):
        result = execute_dagster_graphql(
            graphql_context,
            RELOAD_REPOSITORY_LOCATION_QUERY,
            {"repositoryLocationName": main_repo_location_name()},
        )

        assert result
        assert result.data
        assert result.data["reloadRepositoryLocation"]
        assert result.data["reloadRepositoryLocation"]["__typename"] == "WorkspaceLocationEntry"
        assert result.data["reloadRepositoryLocation"]["name"] == main_repo_location_name()
        repositories = result.data["reloadRepositoryLocation"]["locationOrLoadError"][
            "repositories"
        ]
        assert len(repositories) == 1
        assert repositories[0]["name"] == "test_repo"

        assert (
            result.data["reloadRepositoryLocation"]["locationOrLoadError"]["isReloadSupported"]
            is True
        )

        with mock.patch(
            # note it where the function is *used* that needs to mocked, not
            # where it is defined.
            # see https://docs.python.org/3/library/unittest.mock.html#where-to-patch
            "dagster._api.list_repositories.sync_list_repositories_grpc"
        ) as cli_command_mock:
            with mock.patch(
                # note it where the function is *used* that needs to mocked, not
                # where it is defined.
                # see https://docs.python.org/3/library/unittest.mock.html#where-to-patch
                "dagster._api.snapshot_repository.sync_get_external_repositories_data_grpc"
            ) as remote_repository_mock:

                @repository
                def new_repo():
                    return []

                new_repo_data = RepositorySnap.from_def(new_repo)

                remote_repository_mock.return_value = {
                    "new_repo": (
                        new_repo_data,
                        {},
                    )
                }

                cli_command_mock.return_value = ListRepositoriesResponse(
                    repository_symbols=[],
                    executable_path=sys.executable,
                    repository_code_pointer_dict={
                        "new_repo": CodePointer.from_python_file(__file__, "new_repo", None)
                    },
                )

                result = execute_dagster_graphql(
                    graphql_context,
                    RELOAD_REPOSITORY_LOCATION_QUERY,
                    {"repositoryLocationName": main_repo_location_name()},
                )

                assert cli_command_mock.call_count == 1
                assert remote_repository_mock.call_count == 1

                repositories = result.data["reloadRepositoryLocation"]["locationOrLoadError"][
                    "repositories"
                ]
                assert len(repositories) == 1
                assert repositories[0]["name"] == "new_repo"

    def test_reload_failure(self, graphql_context):
        result = execute_dagster_graphql(
            graphql_context,
            RELOAD_REPOSITORY_LOCATION_QUERY,
            {"repositoryLocationName": main_repo_location_name()},
        )

        assert result
        assert result.data
        assert result.data["reloadRepositoryLocation"]
        assert (
            result.data["reloadRepositoryLocation"]["locationOrLoadError"]["__typename"]
            == "RepositoryLocation"
        )
        assert result.data["reloadRepositoryLocation"]["name"] == main_repo_location_name()
        repositories = result.data["reloadRepositoryLocation"]["locationOrLoadError"][
            "repositories"
        ]
        assert len(repositories) == 1
        assert repositories[0]["name"] == "test_repo"
        assert (
            result.data["reloadRepositoryLocation"]["locationOrLoadError"]["isReloadSupported"]
            is True
        )

        with mock.patch(
            # note it where the function is *used* that needs to mocked, not
            # where it is defined.
            # see https://docs.python.org/3/library/unittest.mock.html#where-to-patch
            "dagster._api.list_repositories.sync_list_repositories_grpc"
        ) as cli_command_mock:
            cli_command_mock.side_effect = Exception("Mocked repository load failure")

            result = execute_dagster_graphql(
                graphql_context,
                RELOAD_REPOSITORY_LOCATION_QUERY,
                {"repositoryLocationName": main_repo_location_name()},
            )

            assert result
            assert result.data
            assert result.data["reloadRepositoryLocation"]
            assert (
                result.data["reloadRepositoryLocation"]["locationOrLoadError"]["__typename"]
                == "PythonError"
            )
            assert result.data["reloadRepositoryLocation"]["name"] == main_repo_location_name()
            assert (
                "Mocked repository load failure"
                in result.data["reloadRepositoryLocation"]["locationOrLoadError"]["message"]
            )

            # Verify failure is idempotent
            result = execute_dagster_graphql(
                graphql_context,
                RELOAD_REPOSITORY_LOCATION_QUERY,
                {"repositoryLocationName": main_repo_location_name()},
            )

            assert result
            assert result.data
            assert result.data["reloadRepositoryLocation"]
            assert (
                result.data["reloadRepositoryLocation"]["locationOrLoadError"]["__typename"]
                == "PythonError"
            )
            assert result.data["reloadRepositoryLocation"]["name"] == main_repo_location_name()
            assert (
                "Mocked repository load failure"
                in result.data["reloadRepositoryLocation"]["locationOrLoadError"]["message"]
            )

        # can be reloaded again successfully
        result = execute_dagster_graphql(
            graphql_context,
            RELOAD_REPOSITORY_LOCATION_QUERY,
            {"repositoryLocationName": main_repo_location_name()},
        )

        assert result
        assert result.data
        assert result.data["reloadRepositoryLocation"]
        assert (
            result.data["reloadRepositoryLocation"]["locationOrLoadError"]["__typename"]
            == "RepositoryLocation"
        )
        assert result.data["reloadRepositoryLocation"]["name"] == main_repo_location_name()
        assert result.data["reloadRepositoryLocation"]["loadStatus"] == "LOADED"
        repositories = result.data["reloadRepositoryLocation"]["locationOrLoadError"][
            "repositories"
        ]
        assert len(repositories) == 1
        assert repositories[0]["name"] == "test_repo"
        assert (
            result.data["reloadRepositoryLocation"]["locationOrLoadError"]["isReloadSupported"]
            is True
        )


class TestReloadRepositoriesManagedGrpc(ManagedTestSuite):
    def test_managed_grpc_reload_location(self, graphql_context):
        result = execute_dagster_graphql(
            graphql_context,
            RELOAD_REPOSITORY_LOCATION_QUERY,
            {"repositoryLocationName": main_repo_location_name()},
        )

        assert result
        assert result.data
        assert result.data["reloadRepositoryLocation"]
        assert (
            result.data["reloadRepositoryLocation"]["locationOrLoadError"]["__typename"]
            == "RepositoryLocation"
        )
        assert result.data["reloadRepositoryLocation"]["name"] == main_repo_location_name()
        assert result.data["reloadRepositoryLocation"]["loadStatus"] == "LOADED"

        repositories = result.data["reloadRepositoryLocation"]["locationOrLoadError"][
            "repositories"
        ]
        assert len(repositories) == 1
        assert repositories[0]["name"] == "test_repo"

        metadatas = repositories[0]["displayMetadata"]
        metadata_dict = {metadata["key"]: metadata["value"] for metadata in metadatas}

        assert (
            "python_file" in metadata_dict
            or "module_name" in metadata_dict
            or "package_name" in metadata_dict
        )

        assert (
            result.data["reloadRepositoryLocation"]["locationOrLoadError"]["isReloadSupported"]
            is True
        )


class TestReloadLocationCodeServerCliGrpc(CodeServerCliTestSuite):
    def test_code_server_cli_reload_location(self, graphql_context):
        # server_id before

        old_server_id = graphql_context.get_code_location(main_repo_location_name()).server_id

        result = execute_dagster_graphql(
            graphql_context,
            RELOAD_REPOSITORY_LOCATION_QUERY,
            {"repositoryLocationName": main_repo_location_name()},
        )

        assert result
        assert result.data
        assert result.data["reloadRepositoryLocation"]
        assert (
            result.data["reloadRepositoryLocation"]["locationOrLoadError"]["__typename"]
            == "RepositoryLocation"
        )
        assert result.data["reloadRepositoryLocation"]["name"] == main_repo_location_name()
        assert result.data["reloadRepositoryLocation"]["loadStatus"] == "LOADED"

        new_location = (
            graphql_context.process_context.get_current_workspace()
            .code_location_entries[main_repo_location_name()]
            .code_location
        )

        assert new_location.server_id != old_server_id  # Reload actually happened



import shutil
from collections.abc import Mapping, Sequence
from pathlib import Path
from typing import Any

import pytest
from dagster import (
    AssetIn,
    OpExecutionContext,
    RepositoryDefinition,
    TimeWindowPartitionMapping,
    asset,
    job,
    op,
    repository,
)
from dagster._config.pythonic_config import Config
from dagster._core.definitions.data_version import DATA_VERSION_TAG, DataVersion
from dagster._core.definitions.decorators.source_asset_decorator import observable_source_asset
from dagster._core.definitions.definitions_class import Definitions
from dagster._core.definitions.events import AssetKey, Output
from dagster._core.definitions.partitions.definition import (
    DailyPartitionsDefinition,
    DynamicPartitionsDefinition,
    StaticPartitionsDefinition,
)
from dagster._core.definitions.unresolved_asset_job_definition import define_asset_job
from dagster._core.test_utils import wait_for_runs_to_finish
from dagster._core.workspace.context import WorkspaceRequestContext
from dagster_graphql.test.utils import (
    define_out_of_process_context,
    ensure_dagster_graphql_tests_import,
    execute_dagster_graphql,
    infer_job_selector,
    infer_repository_selector,
    materialize_assets,
    temp_workspace_file,
)

ensure_dagster_graphql_tests_import()

from dagster_graphql_tests.graphql.graphql_context_test_suite import (
    ExecutingGraphQLContextTestMatrix,
)
from dagster_graphql_tests.graphql.test_assets import (
    GET_ASSET_DATA_VERSIONS,
    GET_ASSET_DATA_VERSIONS_BY_PARTITION,
    GET_ASSET_JOB_NAMES,
)


class TestDataVersions(ExecutingGraphQLContextTestMatrix):
    # Storage dir is shared between tests in the same class, so we need to clean it out.
    @pytest.fixture(autouse=True)
    def clean_storage_directory(self, graphql_context):
        """Clean IO manager storage directory before each test to prevent pollution."""
        # Clean before test runs
        storage_dir = Path(graphql_context.instance.storage_directory())
        if storage_dir.exists():
            shutil.rmtree(storage_dir)
            storage_dir.mkdir(parents=True)
        yield
        # Clean after test runs
        if storage_dir.exists():
            shutil.rmtree(storage_dir)
            storage_dir.mkdir(parents=True)

    def test_dependencies_changed(self, graphql_context):
        repo_v2 = get_repo_v2()
        instance = graphql_context.instance

        with define_out_of_process_context(__file__, "get_repo_v1", instance) as context_v1:
            assert materialize_assets(context_v1, [AssetKey(["foo"]), AssetKey(["bar"])])
            wait_for_runs_to_finish(context_v1.instance)
        with define_out_of_process_context(__file__, "get_repo_v2", instance) as context_v2:
            assert _fetch_data_versions(context_v2, repo_v2)

    def test_stale_status(self, graphql_context):
        repo = get_repo_v1()
        instance = graphql_context.instance

        with define_out_of_process_context(__file__, "get_repo_v1", instance) as context:
            result = _fetch_data_versions(context, repo)
            foo = _get_asset_node(result, "foo")
            assert foo["dataVersion"] is None
            assert foo["staleStatus"] == "MISSING"
            assert foo["staleCauses"] == []

            assert materialize_assets(context, [AssetKey(["foo"]), AssetKey(["bar"])])
            wait_for_runs_to_finish(context.instance)

        with define_out_of_process_context(__file__, "get_repo_v1", instance) as context:
            result = _fetch_data_versions(context, repo)
            foo = _get_asset_node(result, "foo")
            assert foo["dataVersion"] is not None
            assert foo["staleStatus"] == "FRESH"
            assert foo["staleCauses"] == []

            assert materialize_assets(context, asset_selection=[AssetKey(["foo"])])
            wait_for_runs_to_finish(context.instance)

        with define_out_of_process_context(__file__, "get_repo_v1", instance) as context:
            result = _fetch_data_versions(context, repo)
            bar = _get_asset_node(result, "bar")
            assert bar["dataVersion"] is not None
            assert bar["staleStatus"] == "STALE"
            assert bar["staleCauses"] == [
                {
                    "key": {"path": ["foo"]},
                    "category": "DATA",
                    "reason": "has a new materialization",
                    "dependency": None,
                }
            ]

    def test_stale_status_partitioned(self, graphql_context):
        instance = graphql_context.instance
        instance.add_dynamic_partitions("dynamic", ["alpha", "beta"])

        with define_out_of_process_context(__file__, "get_repo_partitioned", instance) as context:
            for key in ["foo", "bar", "dynamic_asset"]:
                result = _fetch_partition_data_versions(context, AssetKey([key]))
                node = _get_asset_node(result)
                assert node["dataVersion"] is None
                assert node["dataVersionByPartition"] == [None, None]
                assert (
                    node["staleStatus"] == "FRESH"
                )  # always returns fresh for this undefined field
                assert node["staleStatusByPartition"] == ["MISSING", "MISSING"]
                assert node["staleCauses"] == []  # always returns empty for this undefined field
                assert node["staleCausesByPartition"] == [[], []]

            assert materialize_assets(
                context,
                [AssetKey(["foo"]), AssetKey(["bar"]), AssetKey("dynamic_asset")],
                ["alpha", "beta"],
            )
            wait_for_runs_to_finish(context.instance)

        with define_out_of_process_context(__file__, "get_repo_partitioned", instance) as context:
            for key in ["foo", "bar", "dynamic_asset"]:
                result = _fetch_partition_data_versions(context, AssetKey([key]), "alpha")
                node = _get_asset_node(result)
                assert node["dataVersion"] == f"ok_{key}_alpha"
                assert node["dataVersionByPartition"] == [f"ok_{key}_alpha", f"ok_{key}_beta"]
                assert node["staleStatus"] == "FRESH"
                assert node["staleStatusByPartition"] == ["FRESH", "FRESH"]
                assert node["staleCauses"] == []
                assert node["staleCausesByPartition"] == [[], []]

            assert materialize_assets(
                context,
                [AssetKey(["foo"])],
                ["alpha", "beta"],
                run_config_data={"ops": {"foo": {"config": {"prefix": "from_config"}}}},
            )
            wait_for_runs_to_finish(context.instance)

        with define_out_of_process_context(__file__, "get_repo_partitioned", instance) as context:
            result = _fetch_partition_data_versions(context, AssetKey(["foo"]), "alpha")
            foo = _get_asset_node(result, "foo")
            assert foo["dataVersion"] == "from_config_foo_alpha"
            assert foo["dataVersionByPartition"] == [
                "from_config_foo_alpha",
                "from_config_foo_beta",
            ]
            assert foo["staleStatus"] == "FRESH"
            assert foo["staleStatusByPartition"] == ["FRESH", "FRESH"]

            result = _fetch_partition_data_versions(
                context, AssetKey(["bar"]), "alpha", ["beta", "alpha"]
            )
            bar = _get_asset_node(result, "bar")
            assert bar["dataVersion"] == "ok_bar_alpha"
            assert bar["dataVersionByPartition"] == ["ok_bar_beta", "ok_bar_alpha"]
            assert bar["staleStatus"] == "STALE"
            assert bar["staleStatusByPartition"] == ["STALE", "STALE"]
            assert bar["staleCauses"] == [
                {
                    "key": {"path": ["foo"]},
                    "partitionKey": "alpha",
                    "category": "DATA",
                    "reason": "has a new data version",
                    "dependency": None,
                    "dependencyPartitionKey": None,
                }
            ]
            assert bar["staleCausesByPartition"] == [
                [
                    {
                        "key": {"path": ["foo"]},
                        "partitionKey": key,
                        "category": "DATA",
                        "reason": "has a new data version",
                        "dependency": None,
                        "dependencyPartitionKey": None,
                    }
                ]
                for key in ["beta", "alpha"]
            ]

    def test_cross_repo_dependency(self, graphql_context):
        instance = graphql_context.instance

        with (
            temp_workspace_file(
                [
                    ("repo1", __file__, "get_cross_repo_test_repo1"),
                    ("repo2", __file__, "get_cross_repo_test_repo2"),
                ]
            ) as workspace_file,
            define_out_of_process_context(workspace_file, None, instance) as context,
        ):
            # Materialize the downstream asset. Provenance will store the version of foo as INITIAL.
            assert materialize_assets(context, [AssetKey(["bar"])], location_name="repo2")
            wait_for_runs_to_finish(context.instance)

            repo2 = get_cross_repo_test_repo2()
            result = _fetch_data_versions(context, repo2, location_name="repo2")
            bar = _get_asset_node(result, "bar")
            assert bar["staleStatus"] == "FRESH"

    def test_data_version_from_tags(self, graphql_context):
        repo_v1 = get_repo_v1()
        instance = graphql_context.instance

        with define_out_of_process_context(__file__, "get_repo_v1", instance) as context_v1:
            assert materialize_assets(context_v1, [AssetKey(["foo"]), AssetKey(["bar"])])
            wait_for_runs_to_finish(context_v1.instance)
            result = _fetch_data_versions(context_v1, repo_v1)
            tags = result.data["assetNodes"][0]["assetMaterializations"][0]["tags"]
            dv_tag = next(tag for tag in tags if tag["key"] == DATA_VERSION_TAG)
            assert dv_tag["value"] == result.data["assetNodes"][0]["dataVersion"]

    def test_partitioned_self_dep(self, graphql_context):
        repo = get_repo_with_partitioned_self_dep_asset()
        instance = graphql_context.instance

        with define_out_of_process_context(
            __file__, "get_repo_with_partitioned_self_dep_asset", instance
        ) as context:
            result = _fetch_partition_data_versions(
                context, AssetKey(["a"]), partitions=["2020-01-01", "2020-01-02"]
            )
            assert result
            assert result.data
            node = _get_asset_node(result, "a")
            assert node["staleStatusByPartition"] == ["MISSING", "MISSING"]

            result = _fetch_data_versions(context, repo)
            node = _get_asset_node(result, "b")
            assert node["staleStatus"] == "MISSING"

    def test_source_asset_job_name(self, graphql_context):
        get_observable_source_asset_repo()
        instance = graphql_context.instance

        with define_out_of_process_context(
            __file__, "get_observable_source_asset_repo", instance
        ) as context:
            selector = infer_repository_selector(context)
            result = execute_dagster_graphql(
                context,
                GET_ASSET_JOB_NAMES,
                variables={
                    "selector": selector,
                },
            )
            assert result and result.data
            foo_jobs = _get_asset_node(result.data["repositoryOrError"], "foo")["jobNames"]
            bar_jobs = _get_asset_node(result.data["repositoryOrError"], "bar")["jobNames"]
            baz_jobs = _get_asset_node(result.data["repositoryOrError"], "baz")["jobNames"]

            # All nodes should have separate job sets since there are two different partitions defs
            # and a non-partitioned observable.
            assert foo_jobs and foo_jobs != bar_jobs and foo_jobs != baz_jobs
            assert bar_jobs and bar_jobs != foo_jobs and bar_jobs != baz_jobs
            assert baz_jobs and baz_jobs != foo_jobs and baz_jobs != bar_jobs

            # Make sure none of our assets included in non-asset job
            assert "bop_job" not in foo_jobs
            assert "bop_job" not in bar_jobs
            assert "bop_job" not in baz_jobs

            # Make sure observable source asset does not appear in non-explicit observation job
            assert "foo_job" not in bar_jobs

            # Make sure observable source asset appears in explicit observation job
            assert "bar_job" in bar_jobs


# ########################
# ##### REPOSITORY DEFS
# ########################

# These are used to define definition state in the test functions above


def get_repo_v1():
    @asset
    def foo():
        return True

    @asset
    def bar(foo):
        return True

    @repository
    def repo():
        return [foo, bar]

    return repo


def get_repo_v2():
    @asset
    def bar():
        return True

    @repository
    def repo():
        return [bar]

    return repo


def get_repo_partitioned():
    partitions_def = StaticPartitionsDefinition(["alpha", "beta"])

    dynamic_partitions_def = DynamicPartitionsDefinition(name="dynamic")

    class FooConfig(Config):
        prefix: str = "ok"

    @asset(partitions_def=partitions_def)
    def foo(context: OpExecutionContext, config: FooConfig) -> Output[bool]:
        return Output(
            True, data_version=DataVersion(f"{config.prefix}_foo_{context.partition_key}")
        )

    @asset(partitions_def=partitions_def)
    def bar(context: OpExecutionContext, foo) -> Output[bool]:
        return Output(True, data_version=DataVersion(f"ok_bar_{context.partition_key}"))

    @asset(partitions_def=dynamic_partitions_def)
    def dynamic_asset(context: OpExecutionContext):
        return Output(True, data_version=DataVersion(f"ok_dynamic_asset_{context.partition_key}"))

    @repository
    def repo():
        return [foo, bar, dynamic_asset]

    return repo


def get_cross_repo_test_repo1():
    @asset
    def foo():
        pass

    @repository
    def repo1():
        return [foo]

    return repo1


def get_cross_repo_test_repo2():
    @asset(deps=["foo"])
    def bar():
        pass

    @repository
    def repo2():
        return [bar]

    return repo2


def get_repo_with_partitioned_self_dep_asset():
    @asset(
        partitions_def=DailyPartitionsDefinition(start_date="2020-01-01"),
        ins={
            "a": AssetIn(
                partition_mapping=TimeWindowPartitionMapping(start_offset=-1, end_offset=-1)
            )
        },
    )
    def a(a):
        del a

    @asset
    def b(a):
        return a

    @repository
    def repo():
        return [a, b]

    return repo


def get_observable_source_asset_repo():
    @asset(partitions_def=StaticPartitionsDefinition(["1"]))
    def foo():
        return 1

    @observable_source_asset
    def bar():
        return DataVersion("1")

    @asset(partitions_def=StaticPartitionsDefinition(["2"]))
    def baz():
        return 1

    @op
    def bop():
        pass

    @job
    def bop_job():
        bop()

    foo_job = define_asset_job("foo_job", [foo])
    bar_job = define_asset_job("bar_job", [bar])

    defs = Definitions(
        assets=[foo, bar, baz],
        jobs=[foo_job, bar_job, bop_job],
    )
    return defs.get_repository_def()


# ########################
# ##### HELPERS
# ########################


def _fetch_data_versions(
    context: WorkspaceRequestContext,
    repo: RepositoryDefinition,
    location_name: str | None = None,
):
    selector = infer_job_selector(
        context, repo.get_implicit_asset_job_names()[0], location_name=location_name
    )
    return execute_dagster_graphql(
        context,
        GET_ASSET_DATA_VERSIONS,
        variables={
            "pipelineSelector": selector,
        },
    )


def _fetch_partition_data_versions(
    context: WorkspaceRequestContext,
    asset_key: AssetKey,
    partition: str | None = None,
    partitions: Sequence[str] | None = None,
):
    return execute_dagster_graphql(
        context,
        GET_ASSET_DATA_VERSIONS_BY_PARTITION,
        variables={
            "assetKey": asset_key.to_graphql_input(),
            "partition": partition,
            "partitions": partitions,
        },
    )


def _get_asset_node(result: Any, key: str | None = None) -> Mapping[str, Any]:
    to_check = result if isinstance(result, dict) else result.data  # GqlResult
    if key is None:  # assume we are dealing with a single-node query
        return to_check["assetNodeOrError"]
    else:
        return (
            to_check["assetNodeOrError"]
            if "assetNodeOrError" in to_check
            else next(node for node in to_check["assetNodes"] if node["assetKey"]["path"] == [key])
        )



from dagster import AssetKey, ConfigurableResource, Definitions, asset, job, op
from dagster._core.definitions.partitions.definition import StaticPartitionsDefinition
from dagster._core.definitions.partitions.partitioned_config import static_partitioned_config
from dagster._core.definitions.repository_definition import SINGLETON_REPOSITORY_NAME
from dagster._core.test_utils import ensure_dagster_tests_import, instance_for_test
from dagster_graphql.test.utils import define_out_of_process_context, execute_dagster_graphql

ensure_dagster_tests_import()


GET_PARTITIONS_QUERY = """
    query SingleJobQuery($selector: PipelineSelector!, $selectedAssetKeys: [AssetKeyInput!]) {
        pipelineOrError(params: $selector) {
            ... on Pipeline {
                name
                partitionKeysOrError(selectedAssetKeys: $selectedAssetKeys) {
                    partitionKeys
                }
            }
        }
    }
"""

GET_PARTITION_TAGS_QUERY = """
    query SingleJobQuery($selector: PipelineSelector!, $partitionName: String!, $selectedAssetKeys: [AssetKeyInput!]) {
        pipelineOrError(params: $selector) {
            ... on Pipeline {
                name
                partition(partitionName: $partitionName, selectedAssetKeys: $selectedAssetKeys) {
                    name
                    tagsOrError {
                        ... on PartitionTags {
                            results {
                                key
                                value
                            }
                        }
                    }
                }
            }
        }
    }
"""

GET_PARTITION_RUN_CONFIG_QUERY = """
    query SingleJobQuery($selector: PipelineSelector!, $partitionName: String!, $selectedAssetKeys: [AssetKeyInput!]) {
        pipelineOrError(params: $selector) {
            ... on Pipeline {
                name
                partition(partitionName: $partitionName, selectedAssetKeys: $selectedAssetKeys) {
                    name
                    runConfigOrError {
                        ... on PartitionRunConfig {
                            yaml
                        }
                    }
                }
            }
        }
    }
"""


def get_repo_with_partitioned_op_job():
    @op
    def op1(): ...

    @static_partitioned_config(["1", "2"])
    def my_partitioned_config(partition):
        return {"ops": {"op1": {"config": {"p": partition}}}}

    @job(config=my_partitioned_config)
    def job1():
        op1()

    return Definitions(jobs=[job1]).get_repository_def()


def get_repo_with_differently_partitioned_assets():
    @asset(partitions_def=StaticPartitionsDefinition(["1", "2"]))
    def asset1(): ...

    ab_partitions_def = StaticPartitionsDefinition(["a", "b"])

    @asset(partitions_def=ab_partitions_def)
    def asset2(): ...

    class MyResource(ConfigurableResource):
        foo: str

    @asset(partitions_def=ab_partitions_def)
    def asset3(resource1: MyResource): ...

    return Definitions(
        assets=[asset1, asset2, asset3], resources={"resource1": MyResource(foo="bar")}
    ).get_repository_def()


def get_repo_with_unpartitioned_assets():
    @asset
    def asset1(): ...

    @asset
    def asset2(): ...

    return Definitions(
        assets=[asset1, asset2],
    ).get_repository_def()


def test_get_partition_names():
    with instance_for_test() as instance:
        with define_out_of_process_context(
            __file__, "get_repo_with_partitioned_op_job", instance
        ) as context:
            result = execute_dagster_graphql(
                context,
                GET_PARTITIONS_QUERY,
                variables={
                    "selector": {
                        "repositoryLocationName": context.code_location_names[0],
                        "repositoryName": SINGLETON_REPOSITORY_NAME,
                        "pipelineName": "job1",
                    }
                },
            )
            assert result.data["pipelineOrError"]["name"] == "job1"
            assert result.data["pipelineOrError"]["partitionKeysOrError"]["partitionKeys"] == [
                "1",
                "2",
            ]


def test_get_partition_names_asset_selection():
    with instance_for_test() as instance:
        with define_out_of_process_context(
            __file__, "get_repo_with_differently_partitioned_assets", instance
        ) as context:
            result = execute_dagster_graphql(
                context,
                GET_PARTITIONS_QUERY,
                variables={
                    "selector": {
                        "repositoryLocationName": context.code_location_names[0],
                        "repositoryName": SINGLETON_REPOSITORY_NAME,
                        "pipelineName": "__ASSET_JOB",
                    },
                    "selectedAssetKeys": [
                        AssetKey("asset2").to_graphql_input(),
                        AssetKey("asset3").to_graphql_input(),
                    ],
                },
            )
            assert result.data["pipelineOrError"]["name"] == "__ASSET_JOB"
            assert result.data["pipelineOrError"]["partitionKeysOrError"]["partitionKeys"] == [
                "a",
                "b",
            ]


def test_get_partition_names_asset_selection_no_partitions():
    with instance_for_test() as instance:
        with define_out_of_process_context(
            __file__, "get_repo_with_unpartitioned_assets", instance
        ) as context:
            result = execute_dagster_graphql(
                context,
                GET_PARTITIONS_QUERY,
                variables={
                    "selector": {
                        "repositoryLocationName": context.code_location_names[0],
                        "repositoryName": SINGLETON_REPOSITORY_NAME,
                        "pipelineName": "__ASSET_JOB",
                    },
                    "selectedAssetKeys": [
                        AssetKey("asset1").to_graphql_input(),
                        AssetKey("asset2").to_graphql_input(),
                    ],
                },
            )
            assert result.data["pipelineOrError"]["name"] == "__ASSET_JOB"
            assert result.data["pipelineOrError"]["partitionKeysOrError"]["partitionKeys"] == []


def test_get_partition_tags():
    with instance_for_test() as instance:
        with define_out_of_process_context(
            __file__, "get_repo_with_partitioned_op_job", instance
        ) as context:
            result = execute_dagster_graphql(
                context,
                GET_PARTITION_TAGS_QUERY,
                variables={
                    "selector": {
                        "repositoryLocationName": context.code_location_names[0],
                        "repositoryName": SINGLETON_REPOSITORY_NAME,
                        "pipelineName": "job1",
                    },
                    "partitionName": "1",
                },
            )
            assert result.data["pipelineOrError"]["name"] == "job1"
            result_partition = result.data["pipelineOrError"]["partition"]
            assert result_partition["name"] == "1"
            assert {
                item["key"]: item["value"] for item in result_partition["tagsOrError"]["results"]
            } == {
                "dagster/partition": "1",
                "dagster/partition_set": "job1_partition_set",
            }


def test_get_partition_tags_asset_selection():
    with instance_for_test() as instance:
        with define_out_of_process_context(
            __file__, "get_repo_with_differently_partitioned_assets", instance
        ) as context:
            result = execute_dagster_graphql(
                context,
                GET_PARTITION_TAGS_QUERY,
                variables={
                    "selector": {
                        "repositoryLocationName": context.code_location_names[0],
                        "repositoryName": SINGLETON_REPOSITORY_NAME,
                        "pipelineName": "__ASSET_JOB",
                    },
                    "selectedAssetKeys": [
                        AssetKey("asset2").to_graphql_input(),
                        AssetKey("asset3").to_graphql_input(),
                    ],
                    "partitionName": "b",
                },
            )
            assert result.data["pipelineOrError"]["name"] == "__ASSET_JOB"
            result_partition = result.data["pipelineOrError"]["partition"]
            assert result_partition["name"] == "b"
            assert {
                item["key"]: item["value"] for item in result_partition["tagsOrError"]["results"]
            } == {"dagster/partition": "b"}


def test_get_partition_config():
    with instance_for_test() as instance:
        with define_out_of_process_context(
            __file__, "get_repo_with_partitioned_op_job", instance
        ) as context:
            result = execute_dagster_graphql(
                context,
                GET_PARTITION_RUN_CONFIG_QUERY,
                variables={
                    "selector": {
                        "repositoryLocationName": context.code_location_names[0],
                        "repositoryName": SINGLETON_REPOSITORY_NAME,
                        "pipelineName": "job1",
                    },
                    "partitionName": "1",
                },
            )
            assert result.data["pipelineOrError"]["name"] == "job1"
            result_partition = result.data["pipelineOrError"]["partition"]
            assert result_partition["name"] == "1"
            assert (
                result_partition["runConfigOrError"]["yaml"]
                == """ops:\n  op1:\n    config:\n      p: '1'\n"""
            )


def test_get_partition_config_asset_selection():
    with instance_for_test() as instance:
        with define_out_of_process_context(
            __file__, "get_repo_with_differently_partitioned_assets", instance
        ) as context:
            result = execute_dagster_graphql(
                context,
                GET_PARTITION_RUN_CONFIG_QUERY,
                variables={
                    "selector": {
                        "repositoryLocationName": context.code_location_names[0],
                        "repositoryName": SINGLETON_REPOSITORY_NAME,
                        "pipelineName": "__ASSET_JOB",
                    },
                    "selectedAssetKeys": [
                        AssetKey("asset2").to_graphql_input(),
                        AssetKey("asset3").to_graphql_input(),
                    ],
                    "partitionName": "b",
                },
            )
            assert result.data["pipelineOrError"]["name"] == "__ASSET_JOB"
            result_partition = result.data["pipelineOrError"]["partition"]
            assert result_partition["name"] == "b"
            assert result_partition["runConfigOrError"]["yaml"] == "{}\n"



from dagster_graphql.test.utils import execute_dagster_graphql, infer_repository_selector

UTILIZED_ENV_VARS_QUERY = """
query UtilizedEnvVarsQuery($selector: RepositorySelector!) {
  utilizedEnvVarsOrError(repositorySelector: $selector) {
    __typename
    ... on EnvVarWithConsumersList{
      results {
        envVarName
        envVarConsumers {
            type
            name
        }
      }
    }
  }
}
"""


def test_get_used_env_vars(definitions_graphql_context, snapshot) -> None:
    selector = infer_repository_selector(definitions_graphql_context)
    result = execute_dagster_graphql(
        definitions_graphql_context,
        UTILIZED_ENV_VARS_QUERY,
        {"selector": selector},
    )
    assert not result.errors
    assert result.data
    assert result.data["utilizedEnvVarsOrError"]

    def sort_env_var_entry(entry):
        return {
            "envVarName": entry["envVarName"],
            "envVarConsumers": sorted(entry["envVarConsumers"], key=lambda x: x["name"]),
        }

    sorted_env_vars = sorted(
        [sort_env_var_entry(x) for x in result.data["utilizedEnvVarsOrError"]["results"]],
        key=lambda x: x["envVarName"],
    )

    assert sorted_env_vars == [
        {
            "envVarName": "MY_OTHER_STRING",
            "envVarConsumers": [
                {
                    "type": "RESOURCE",
                    "name": "my_resource_two_env_vars",
                }
            ],
        },
        {
            "envVarName": "MY_STRING",
            "envVarConsumers": [
                {
                    "type": "RESOURCE",
                    "name": "my_resource_env_vars",
                },
                {
                    "type": "RESOURCE",
                    "name": "my_resource_two_env_vars",
                },
            ],
        },
    ]

    result.data["utilizedEnvVarsOrError"]["results"] = sorted_env_vars

    snapshot.assert_match(result.data)
